# type: ignore

"""
https://github.com/WolfgangFahl/pyLoDStorage/issues/127
https://docs.python.org/3/library/sqlite3.html#adapter-and-converter-recipes

fix: DeprecationWarning: The default date converter is deprecated as of Python 3.12;
 see the sqlite3 documentation for suggested replacement recipes
"""

import datetime
import sqlite3


def adapt_date_iso(val):
    """Adapt datetime.date to ISO 8601 date."""
    return val.isoformat()


def adapt_datetime_iso(val):
    """Adapt datetime.datetime to timezone-naive ISO 8601 date."""
    return val.isoformat()


def adapt_datetime_epoch(val):
    """Adapt datetime.datetime to Unix timestamp."""
    return int(val.timestamp())


sqlite3.register_adapter(datetime.date, adapt_date_iso)
sqlite3.register_adapter(datetime.datetime, adapt_datetime_iso)
sqlite3.register_adapter(datetime.datetime, adapt_datetime_epoch)


def convert_date(val):
    """Convert ISO 8601 date to datetime.date object."""
    return datetime.date.fromisoformat(val.decode())


def convert_datetime(val):
    """Convert ISO 8601 datetime to datetime.datetime object."""
    return datetime.datetime.fromisoformat(val.decode())


def convert_timestamp(val):
    """Convert Unix epoch timestamp to datetime.datetime object."""
    return datetime.datetime.fromtimestamp(int(val))  # noqa: DTZ006


sqlite3.register_converter('date', convert_date)
sqlite3.register_converter('datetime', convert_datetime)
sqlite3.register_converter('timestamp', convert_timestamp)



from __future__ import annotations

import re
import statistics
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any

import pytest
from django.db import connection
from django.utils import timezone
from src.capture_db_queries.decorators import CaptureQueries, ExtCaptureQueriesContext, capture_queries
from src.capture_db_queries.wrappers import BaseExecutionWrapper, ExplainExecutionWrapper

from tests.conftest import slow_down_execute
from tests.models import Article, Reporter

if TYPE_CHECKING:
    from collections.abc import Callable, Generator
    from datetime import date
    from io import StringIO
    from uuid import UUID

    from django.db.backends.utils import CursorWrapper

ANYNUM = r'0.0[0-9]+'


def request_to_db(date_now: date | None = None) -> tuple[Reporter, Article]:
    reporter = Reporter.objects.create(full_name='full_name')
    return reporter, Article.objects.create(
        pub_date=date_now or timezone.now().date(),
        headline='headline',
        content='content',
        reporter=reporter,
    )


def _select(reporter_id: UUID, article_id: UUID) -> None:
    list(Reporter.objects.filter(pk=reporter_id))
    list(Article.objects.filter(pk=article_id))


class BasicTestsFor3ChoicesOfCaptureQueries:
    """"""

    def setup_method(self, method: Callable[..., Any]) -> None:
        self.reporter, self.article = request_to_db()

    def call_capture_queries(self, **kwargs: Any) -> CaptureQueries:
        raise NotImplementedError

    # @pytest.mark.usefixtures('_debug_true')
    def test_basic_logic(self) -> None:
        obj = self.call_capture_queries()

        assert isinstance(obj.wrapper, BaseExecutionWrapper), type(obj.wrapper).__name__

        data = (obj.current_iteration, len(obj), obj.connection.vendor)
        assert data == (1, 2, 'sqlite'), data

        data = sum(obj.wrapper.timer.all_execution_times)
        gt, lt = 0.1, 0.3
        assert gt < data < lt, f'{gt} < {data} < {lt}'

        data = statistics.median(obj.wrapper.timer.all_execution_times)
        gt, lt = 0.08, 0.13
        assert gt < data < lt, f'{gt} < {data} < {lt}'

        data = obj.queries_log[0]['time']
        assert gt < data < lt, f'{gt} < {data} < {lt}'

        data = obj.queries_log[1]['time']
        assert gt < data < lt, f'{gt} < {data} < {lt}'

        data = obj.queries_log[0]['sql']
        assert obj.queries_log[0]['sql'] == (
            'SELECT "tests_reporter"."id", "tests_reporter"."full_name" '
            'FROM "tests_reporter" WHERE "tests_reporter"."id" = %s' % self.reporter.pk
        ), data

        data = obj.queries_log[1]['sql']
        assert obj.queries_log[1]['sql'] == (
            'SELECT "tests_article"."id", "tests_article"."pub_date", "tests_article"."headline", '
            '"tests_article"."content", "tests_article"."reporter_id" '
            'FROM "tests_article" WHERE "tests_article"."id" = %s' % self.article.pk
        ), data

        with pytest.raises(KeyError, match='explain'):
            obj.queries_log[0]['explain']

        with pytest.raises(KeyError, match='explain'):
            obj.queries_log[1]['explain']

    def test_param__assert_q_count(self) -> None:
        with pytest.raises(AssertionError, match='2 not less than or equal to 1 queries'):
            self.call_capture_queries(assert_q_count=1)

    def test_param__explain(self) -> None:
        obj = self.call_capture_queries(explain=True)

        assert isinstance(obj.wrapper, ExplainExecutionWrapper), type(obj.wrapper).__name__

        gt, lt = 0.08, 0.13

        data = obj.queries_log[0]['time']
        assert gt < data < lt, f'{gt} < {data} < {lt}'

        data = obj.queries_log[1]['time']
        assert gt < data < lt, f'{gt} < {data} < {lt}'

        data = obj.queries_log[0]['sql']
        assert data == (
            'SELECT "tests_reporter"."id", "tests_reporter"."full_name" '
            'FROM "tests_reporter" WHERE "tests_reporter"."id" = %s' % self.reporter.pk
        ), data

        data = obj.queries_log[1]['sql']
        assert data == (
            'SELECT "tests_article"."id", "tests_article"."pub_date", "tests_article"."headline", '
            '"tests_article"."content", "tests_article"."reporter_id" '
            'FROM "tests_article" WHERE "tests_article"."id" = %s' % self.article.pk
        ), data

        # data = obj.queries_log[0]['explain']
        # assert data == '2 0 0 SEARCH TABLE tests_reporter USING INTEGER PRIMARY KEY (rowid=?)', data

        # data = obj.queries_log[1]['explain']
        # assert data == '2 0 0 SEARCH TABLE tests_article USING INTEGER PRIMARY KEY (rowid=?)', data

    # def test_param__explain_opts(self) -> None:
    #     with pytest.raises(ValueError, match='Unknown options: opt1, opt12'):
    #         self.call_capture_queries(explain=True, explain_opts={'opt1': 'value1', 'opt2': 'value2'})

    def test_param__connection(self) -> None:
        class FakeConnection:
            vendor = 'fake_vendor'
            queries_limit = 4
            ops = connection.ops

            @contextmanager
            def execute_wrapper(
                self, wrapper: BaseExecutionWrapper
            ) -> Generator[None, None, CursorWrapper]:
                connection.execute_wrappers.append(wrapper)
                try:
                    yield
                finally:
                    connection.execute_wrappers.pop()

        obj = self.call_capture_queries(connection=FakeConnection())

        data = obj.connection.vendor
        assert data == FakeConnection.vendor, data


@pytest.mark.django_db(transaction=True)
class TestLoopCaptureQueries(BasicTestsFor3ChoicesOfCaptureQueries):
    """"""

    def call_capture_queries(self, **kwargs: Any) -> CaptureQueries:
        with slow_down_execute(0.1):
            obj = CaptureQueries(**kwargs)
            for _ in obj:
                _select(self.reporter.pk, self.article.pk)
        return obj

    # @pytest.mark.usefixtures('_debug_true')
    def test_param__number_runs(self) -> None:
        obj = self.call_capture_queries(number_runs=3)

        data = (obj.current_iteration, len(obj), obj.connection.vendor)
        assert data == (3, 6, 'sqlite'), data

        data = sum(obj.wrapper.timer.all_execution_times)
        gt, lt = 0.6, 0.7
        assert gt < data < lt, f'{gt} < {data} < {lt}'

        data = statistics.median(obj.wrapper.timer.all_execution_times)
        gt, lt = 0.08, 0.13
        assert gt < data < lt, f'{gt} < {data} < {lt}'

    def test_execute_raw_sql(self) -> None:
        reporter_raw_sql = (
            'SELECT "tests_reporter"."id", "tests_reporter"."full_name" '
            'FROM "tests_reporter" WHERE "tests_reporter"."id" = %s' % self.reporter.pk
        )
        article_raw_sql = (
            'SELECT "tests_article"."id", "tests_article"."pub_date", "tests_article"."headline", '
            '"tests_article"."content", "tests_article"."reporter_id" '
            'FROM "tests_article" WHERE "tests_article"."id" = %s' % self.article.pk
        )
        obj = CaptureQueries()
        for _ in obj:
            list(Reporter.objects.raw(reporter_raw_sql))
            list(Article.objects.raw(article_raw_sql))

        data = obj.queries_log[0]['sql']
        assert data == (reporter_raw_sql), data

        data = obj.queries_log[1]['sql']
        assert data == (article_raw_sql), data

    def test_without_requests(self) -> None:
        for _ in CaptureQueries(advanced_verb=True):
            pass  # no have requests


@pytest.mark.django_db(transaction=True)
class TestDecoratorCaptureQueries(BasicTestsFor3ChoicesOfCaptureQueries):
    """"""

    def call_capture_queries(self, **kwargs: Any) -> CaptureQueries:
        with slow_down_execute(0.1):
            obj = CaptureQueries(**dict({'auto_call_func': True}, **kwargs))

            @obj
            def _() -> None:
                _select(self.reporter.pk, self.article.pk)

        return obj

    def test_param__number_runs(self) -> None:
        with slow_down_execute(0.1):
            obj = CaptureQueries(auto_call_func=False, number_runs=3)

            @obj
            def func(a: int, b: int) -> int:
                _select(self.reporter.pk, self.article.pk)
                return a + b

            result, eq = func(2, 2), 4

        assert result == eq, f'{result} == {eq}'

        data = (obj.current_iteration, len(obj), obj.connection.vendor)
        assert data == (3, 6, 'sqlite'), data

        data = sum(obj.wrapper.timer.all_execution_times)
        gt, lt = 0.6, 0.7
        assert gt < data < lt, f'{gt} < {data} < {lt}'

        data = statistics.median(obj.wrapper.timer.all_execution_times)
        gt, lt = 0.08, 0.13
        assert gt < data < lt, f'{gt} < {data} < {lt}'

    def test_param__auto_call_func(self) -> None:
        obj = self.call_capture_queries(auto_call_func=True)

        data = (obj.current_iteration, len(obj), obj.connection.vendor)
        assert data == (1, 2, 'sqlite'), data

        data = sum(obj.wrapper.timer.all_execution_times)
        gt, lt = 0.1, 0.3
        assert gt < data < lt, f'{gt} < {data} < {lt}'

        data = statistics.median(obj.wrapper.timer.all_execution_times)
        gt, lt = 0.08, 0.13
        assert gt < data < lt, f'{gt} < {data} < {lt}'

    def test_execute_raw_sql(self) -> None:
        reporter_raw_sql = (
            'SELECT "tests_reporter"."id", "tests_reporter"."full_name" '
            'FROM "tests_reporter" WHERE "tests_reporter"."id" = %s' % self.reporter.pk
        )
        article_raw_sql = (
            'SELECT "tests_article"."id", "tests_article"."pub_date", "tests_article"."headline", '
            '"tests_article"."content", "tests_article"."reporter_id" '
            'FROM "tests_article" WHERE "tests_article"."id" = %s' % self.article.pk
        )
        obj = CaptureQueries(auto_call_func=True)

        @obj
        def _() -> None:
            list(Reporter.objects.raw(reporter_raw_sql))
            list(Article.objects.raw(article_raw_sql))

        data = obj.queries_log[0]['sql']
        assert data == (reporter_raw_sql), data

        data = obj.queries_log[1]['sql']
        assert data == (article_raw_sql), data

    def test_without_requests(self) -> None:
        @CaptureQueries(advanced_verb=True, auto_call_func=True)
        def func() -> None:
            pass  # no have requests


@pytest.mark.django_db(transaction=True)
class TestContextManagerCaptureQueries(BasicTestsFor3ChoicesOfCaptureQueries):
    """"""

    def call_capture_queries(self, **kwargs: Any) -> CaptureQueries:
        with slow_down_execute(0.1):  # noqa: SIM117
            with CaptureQueries(**kwargs) as obj:
                _select(self.reporter.pk, self.article.pk)
        return obj

    # @pytest.mark.filterwarnings("ignore::UserWarning")  # warn not show, and not raise exc
    # @pytest.mark.filterwarnings('default::UserWarning')  # warn show, and not raise exc
    def test_param__number_runs(self) -> None:
        with pytest.raises(  # noqa: SIM117
            UserWarning,
            match=(
                'When using: CaptureQueries as a context manager,'
                ' the number_runs > 1 parameter is not used.'
            ),
        ):
            with CaptureQueries(number_runs=3):
                pass

    def test_execute_raw_sql(self) -> None:
        reporter_raw_sql = (
            'SELECT "tests_reporter"."id", "tests_reporter"."full_name" '
            'FROM "tests_reporter" WHERE "tests_reporter"."id" = %s' % self.reporter.pk
        )
        article_raw_sql = (
            'SELECT "tests_article"."id", "tests_article"."pub_date", "tests_article"."headline", '
            '"tests_article"."content", "tests_article"."reporter_id" '
            'FROM "tests_article" WHERE "tests_article"."id" = %s' % self.article.pk
        )
        with CaptureQueries() as obj:
            list(Reporter.objects.raw(reporter_raw_sql))
            list(Article.objects.raw(article_raw_sql))

        data = obj.queries_log[0]['sql']
        assert data == (reporter_raw_sql), data

        data = obj.queries_log[1]['sql']
        assert data == (article_raw_sql), data

    def test_without_requests(self) -> None:
        with CaptureQueries(advanced_verb=True):
            pass  # no have requests


@pytest.mark.django_db(transaction=True)
class TestOutputCaptureQueries:
    """The -s argument must be passed when running py test to output output"""

    # @pytest.mark.usefixtures('_debug_true')
    def test_capture_queries_loop(self, intercept_output: StringIO) -> None:
        date_now = timezone.now().date()

        for _ctx in CaptureQueries(assert_q_count=200, number_runs=100, auto_call_func=True):
            request_to_db(date_now)

        output = intercept_output.getvalue()

        assert re.match(
            f'\n\nTests count: 100  |  Total queries count: 200  |  Total execution time: {ANYNUM}s  |  Median time one test is: {ANYNUM}s\n',  # noqa: E501
            output,
        ), f'incorrect output = {output}'

    # @pytest.mark.usefixtures('_debug_true')
    def test_capture_queries_decorator(self, intercept_output: StringIO) -> None:
        date_now = timezone.now().date()

        @CaptureQueries(assert_q_count=200, number_runs=100, auto_call_func=True)
        def _() -> None:
            request_to_db(date_now)

        output = intercept_output.getvalue()

        assert re.match(
            f'\n\nTests count: 100  |  Total queries count: 200  |  Total execution time: {ANYNUM}s  |  Median time one test is: {ANYNUM}s\n',  # noqa: E501
            output,
        ), f'incorrect output = {output}'

    # @pytest.mark.usefixtures('_debug_true')
    def test_capture_queries_context_manager(self, intercept_output: StringIO) -> None:
        with CaptureQueries() as ctx:  # noqa: F841
            request_to_db()

        output = intercept_output.getvalue()

        assert re.match(
            f'\nQueries count: 2  |  Execution time: {ANYNUM}s  |  Vendor: sqlite\n\n',
            output,
        ), f'incorrect output = {output}'


@pytest.mark.django_db(transaction=True)
@pytest.mark.usefixtures('_ignore_deprecation')
def test_capture_queries(intercept_output: StringIO) -> None:
    """Обязательно должен быть передан аргумент -s при запуске pytest, для вывода output"""

    @capture_queries(assert_q_count=200, number_runs=100)
    def _() -> None:
        request_to_db()

    output = intercept_output.getvalue()

    assert re.match(
        f'Tests count: 100  |  Total queries count: 200  |  Total execution time: {ANYNUM}s  |  Median time one test is: {ANYNUM}s\n',  # noqa: E501
        output,
    ), 'incorrect output'


@pytest.mark.django_db(transaction=True)
@pytest.mark.usefixtures('_ignore_deprecation')
def test_capture_queries_with_advanced_verb_and_queries(intercept_output: StringIO) -> None:
    """Обязательно должен быть передан аргумент -s при запуске pytest, для вывода output"""

    date_now = timezone.now().date()

    @capture_queries(number_runs=2, advanced_verb=True, queries=True)
    def _() -> None:
        request_to_db(date_now)

    output = intercept_output.getvalue()
    result_outputs = [x for x in output.split('\n') if x != '']

    correct_outputs = [
        f'Test №1 | Queries count: 2 | Execution time: {ANYNUM}s',
        f'Test №2 | Queries count: 2 | Execution time: {ANYNUM}s',
        '---SQL query---',
        '---SQL query---',
        '---SQL query---',
        '---SQL query---',
        f'Tests count: 2  |  Total queries count: 4  |  Total execution time: {ANYNUM}s  |  Median time one test is: {ANYNUM}s',  # noqa: E501
    ]

    assert len(result_outputs) == len(correct_outputs)

    for result_output, correct_output in zip(result_outputs, correct_outputs):  # noqa: B905
        if result_output.startswith("{'sql':"):
            continue
        assert re.match(
            correct_output,
            result_output,
        ), 'incorrect output'


@pytest.mark.django_db(transaction=True)
@pytest.mark.usefixtures('_ignore_deprecation')
def test_ext_capture_queries_context(intercept_output: StringIO) -> None:
    """Обязательно должен быть передан аргумент -s при запуске pytest, для вывода output"""

    with ExtCaptureQueriesContext() as ctx:  # noqa: F841
        request_to_db()

    output = intercept_output.getvalue()

    assert re.match(
        f'Queries count: 2  |  Execution time: {ANYNUM}s',
        output,
    ), 'incorrect output'



import warnings

warnings.filterwarnings(
    'ignore',
    message="'cgi' is deprecated and slated for removal in Python 3.13",
    category=DeprecationWarning,
)

from . import _sqlite3_adapters_and_converters  # noqa: F401, E402
from ._logging import switch_logger, switch_trace  # noqa: F401, E402
from .decorators import CaptureQueries, ExtCaptureQueriesContext, capture_queries  # noqa: F401, E402



from __future__ import annotations

import sys
import time
import warnings
from contextlib import contextmanager
from io import StringIO
from typing import TYPE_CHECKING, Any
from unittest.mock import patch

import django
import pytest
from django.conf import settings
from django.db.backends.utils import CursorWrapper
from src.capture_db_queries._logging import switch_logger, switch_trace  # noqa: F401

if TYPE_CHECKING:
    from collections.abc import Callable, Generator


def pytest_configure(config: pytest.Config) -> None:
    settings.configure(
        INSTALLED_APPS=('tests',),
        DATABASES={
            'default': {
                'ENGINE': 'django.db.backends.sqlite3',
                'NAME': ':memory:',
            }
        },
    )
    django.setup()


@contextmanager
def intercept_output_ctx() -> Generator[StringIO, None, None]:
    """Перехватывает вывод для последующей проверки в тесте"""
    capture_output = StringIO()
    sys.stdout = capture_output
    try:
        yield capture_output
    finally:
        sys.stdout = sys.__stdout__


@pytest.fixture
def intercept_output() -> Generator[StringIO, None, None]:
    """Перехватывает вывод для последующей проверки в тесте"""
    with intercept_output_ctx() as capture_output:
        yield capture_output


@pytest.fixture
def _ignore_deprecation() -> Generator[None, None, None]:
    warnings.simplefilter('ignore', DeprecationWarning)
    yield
    warnings.resetwarnings()


@pytest.fixture
def _debug_true() -> Generator[None, None, None]:
    switch_logger(True)
    # switch_trace(True)
    yield
    switch_logger(False)
    # switch_trace(False)


@contextmanager
def slow_down_execute(seconds: float = 0.1) -> Generator[None, None, None]:
    """
    Фикстура для замедления CursorWrapper._execute и CursorWrapper._executemany

    Замедляет как для explain вызова так и для обычного,
    но т.к. время для explain не учитывается, замедления не видно.
    """
    # Сохраняем оригинальные методы _execute и _executemany
    original_execute: Callable[..., CursorWrapper | None] = CursorWrapper._execute
    original_executemany: Callable[..., CursorWrapper | None] = CursorWrapper._executemany

    def slow_execute(
        self: CursorWrapper, sql: str, params: tuple[Any, ...], *ignored_wrapper_args: tuple[Any, ...]
    ) -> CursorWrapper | None:
        time.sleep(seconds)  # Искусственная задержка
        return original_execute(self, sql, params, *ignored_wrapper_args)

    def slow_executemany(
        self: CursorWrapper, sql: str, params: tuple[Any, ...], *ignored_wrapper_args: tuple[Any, ...]
    ) -> CursorWrapper | None:
        time.sleep(seconds)  # Искусственная задержка
        return original_executemany(self, sql, params, *ignored_wrapper_args)

    with patch('django.db.backends.utils.CursorWrapper._execute', new=slow_execute):  # noqa: SIM117
        with patch('django.db.backends.utils.CursorWrapper._executemany', new=slow_executemany):
            yield


@contextmanager
def skip_colorize_sql() -> Generator[None, None, None]:
    with patch(
        'src.capture_db_queries.printers.PrinterSql.colorize_sql', new=lambda _, queries_log: queries_log
    ):
        yield



from __future__ import annotations

import abc
from collections import deque
from typing import TYPE_CHECKING, Any
from unittest.util import safe_repr

import pygments.formatters
import pygments.lexers
import sqlparse

if TYPE_CHECKING:
    from .dtos import IterationPrintDTO, SeveralPrintDTO, SinglePrintDTO
    from .types import QueriesLog

_EXCLUDE = True


class AbcPrinter(abc.ABC):
    def __init__(
        self,
        vendor: str,
        assert_q_count: int | None = None,
        verbose: bool = True,
        advanced_verb: bool = False,
        queries: bool = False,
    ) -> None:
        self.vendor = vendor
        self.assert_q_count = assert_q_count
        self.verbose = verbose
        self.advanced_verb = advanced_verb
        self.queries = queries

    def print_sql(self, template: str, queries_log: QueriesLog, **format_kwargs: Any) -> str:
        if self.queries:
            print(self._beautiful_queries(queries_log))

        if self.verbose:
            data = template.format(**format_kwargs, vendor=self.vendor)
            print(data)
            return data
        return ''

    def _beautiful_queries(self, queries_log: QueriesLog) -> str:
        """"""
        filtered_queries = self.filter_queries(queries_log)

        formatted_queries = self.format_sql(filtered_queries)

        formatted_queries = self.colorize_sql(formatted_queries)

        formatted_queries = self.format_explain(formatted_queries)

        return self.build_output_string(formatted_queries)

    @abc.abstractmethod
    def print_single_sql(self, dto: SinglePrintDTO) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def print_several_sql(self, dto: SeveralPrintDTO) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def iteration_print(self, dto: IterationPrintDTO) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def assert_msg(self, queries_count: int) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def filter_queries(self, queries_log: QueriesLog) -> QueriesLog:
        raise NotImplementedError

    @abc.abstractmethod
    def format_sql(self, queries_log: QueriesLog) -> QueriesLog:
        raise NotImplementedError

    @abc.abstractmethod
    def colorize_sql(self, queries_log: QueriesLog) -> QueriesLog:
        raise NotImplementedError

    @abc.abstractmethod
    def format_explain(self, queries_log: QueriesLog) -> QueriesLog:
        raise NotImplementedError

    @abc.abstractmethod
    def build_output_string(self, queries_log: QueriesLog) -> str:
        raise NotImplementedError


class PrinterSql(AbcPrinter):
    EXCLUDE = ('BEGIN', 'COMMIT', 'ROLLBACK')

    single_sql_template = (
        '\nQueries count: {queries_count}  |  '
        'Execution time: {execution_time_per_iter:.6f}s  |  Vendor: {vendor}\n'
    )
    several_sql_template = (
        'Tests count: {current_iteration}  |  '
        'Total queries count: {queries_count}  |  '
        'Total execution time: {sum_all_execution_times:.5f}s  |  '
        'Median time one test is: {median_all_execution_times:.6f}s  |  '
        'Vendor: {vendor}\n'
    )
    iteration_sql_template = (
        'Test №{current_iteration} | '
        'Queries count: {queries_count_per_iter} | '
        'Execution time: {execution_time_per_iter:.6f}s'
    )
    assert_msg_template = '{queries_count} not less than or equal to {assert_q_count} queries'

    sql_template = '№[{ordinal_num}] time=[{time:.6f}]{explain}\n{sql}'

    def print_single_sql(self, dto: SinglePrintDTO) -> str:
        return self.print_sql(
            self.single_sql_template,
            dto.queries_log,
            queries_count=dto.queries_count,
            execution_time_per_iter=dto.execution_time_per_iter,
        )

    def print_several_sql(self, dto: SeveralPrintDTO) -> str:
        format_kwargs: dict[str, int | float] = {
            'queries_count': dto.queries_count,
            'current_iteration': dto.current_iteration,
            'sum_all_execution_times': dto.sum_all_execution_times,
            'median_all_execution_times': dto.median_all_execution_times,
        }
        if self.verbose and not self.advanced_verb and not self.queries:
            print('\n')
        return self.print_sql(self.several_sql_template, dto.queries_log, **format_kwargs)

    def iteration_print(self, dto: IterationPrintDTO) -> None:
        if self.advanced_verb:
            if dto.current_iteration == 1:
                print('\n')
            print(
                self.iteration_sql_template.format(
                    current_iteration=dto.current_iteration,
                    queries_count_per_iter=dto.queries_count_per_iter,
                    execution_time_per_iter=dto.execution_time_per_iter,
                )
            )

    def assert_msg(self, queries_count: int) -> str:
        return self.assert_msg_template.format(
            queries_count=safe_repr(queries_count), assert_q_count=safe_repr(self.assert_q_count)
        )

    def filter_queries(self, queries_log: QueriesLog) -> QueriesLog:
        if _EXCLUDE:
            return deque(query for query in queries_log if query['sql'].upper() not in self.EXCLUDE)
        return queries_log

    def format_sql(self, queries_log: QueriesLog) -> QueriesLog:
        formatted_queries = queries_log.copy()
        for query in formatted_queries:
            # A hint on the parameters: sqlparse.formatter.validate_options
            query['sql'] = sqlparse.format(
                sql=query['sql'],
                encoding='utf-8',
                output_format='sql',
                reindent_aligned=True,
            )
        return formatted_queries

    def colorize_sql(self, queries_log: QueriesLog) -> QueriesLog:
        for query in queries_log:
            colorized_sql = pygments.highlight(
                query['sql'],
                pygments.lexers.get_lexer_by_name('sql'),
                pygments.formatters.TerminalFormatter(),
            )
            query['sql'] = colorized_sql
        return queries_log

    def format_explain(self, queries_log: QueriesLog) -> QueriesLog:
        for query in queries_log:
            if 'explain' in query:
                explain_count = len(query['explain'].split('\n'))
                if explain_count > 1:
                    explain = f' explain=[\n{query["explain"]}\n]'
                else:
                    explain = f' explain=[{query["explain"]}]'
                query['explain'] = explain
        return queries_log

    def build_output_string(self, queries_log: QueriesLog) -> str:
        formatted_queries = '\n\n\n'.join(
            self.sql_template.format(
                ordinal_num=ordinal_num,
                time=query['time'],
                sql=query['sql'],
                explain=query.get('explain', ''),
            )
            for ordinal_num, query in enumerate(queries_log, start=1)
        )
        return f'\n\n{formatted_queries}\n\n'



from __future__ import annotations

import statistics
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .types import QueriesLog


@dataclass(frozen=True)
class BasePrintDTO:
    queries_count: int
    queries_log: QueriesLog


@dataclass(frozen=True)
class SinglePrintDTO(BasePrintDTO):
    execution_time_per_iter: float


@dataclass(frozen=True)
class SeveralPrintDTO(BasePrintDTO):
    current_iteration: int
    all_execution_times: list[float]

    @property
    def sum_all_execution_times(self) -> float:
        if self.all_execution_times:
            return sum(self.all_execution_times)
        return 0.0

    @property
    def median_all_execution_times(self) -> float:
        if self.all_execution_times:
            return statistics.median(self.all_execution_times)
        return 0.0


@dataclass(frozen=True)
class IterationPrintDTO:
    current_iteration: int
    queries_count_per_iter: int
    execution_time_per_iter: float



from __future__ import annotations

from collections import deque
from collections.abc import Callable
from typing import Any, TypeVar, Union

# https://mypy.readthedocs.io/en/stable/generics.html#decorator-factories
DecoratedCallable = TypeVar('DecoratedCallable', bound=Callable[..., Any])

T = TypeVar('T')

Query = dict[str, Union[str, float]]  # noqa: UP007

QueriesLog = deque[Query]

Explain = list[tuple[Union[int, str], ...]]  # noqa: UP007



from __future__ import annotations

from collections import deque
from typing import TYPE_CHECKING, Any

import pytest
from src.capture_db_queries.dtos import IterationPrintDTO, SeveralPrintDTO, SinglePrintDTO
from src.capture_db_queries.printers import PrinterSql

from tests.conftest import intercept_output_ctx, skip_colorize_sql

if TYPE_CHECKING:
    from collections.abc import Callable

    from src.capture_db_queries.types import QueriesLog


@pytest.mark.django_db(transaction=True)
class TestPrinter:
    """"""

    def setup_method(self, method: Callable[..., Any]) -> None:
        self.skip_colorize_sql = skip_colorize_sql()
        self.skip_colorize_sql.__enter__()

        queries_log: QueriesLog = deque(
            [
                {
                    'sql': 'SELECT "tests_reporter"."id", "tests_reporter"."full_name" '
                    'FROM "tests_reporter" WHERE "tests_reporter"."id" = %s',
                    'time': 0.09399999992456287,
                },
                {
                    'sql': 'SELECT "tests_article"."id", "tests_article"."pub_date", '
                    '"tests_article"."headline", "tests_article"."content", "tests_article"."reporter_id" '  # noqa: E501
                    'FROM "tests_article" WHERE "tests_article"."id" = %s',
                    'time': 0.10900000005494803,
                },
            ],
            maxlen=9000,
        )
        self.queries_log_output = (
            '\n\n№[1] time=[0.094000]\nSELECT "tests_reporter"."id",\n       '
            '"tests_reporter"."full_name"\n  FROM "tests_reporter"\n '
            'WHERE "tests_reporter"."id" = %s\n\n\n№[2] time=[0.109000]\nSELECT "tests_article"."id",\n       '  # noqa: E501
            '"tests_article"."pub_date",\n       '
            '"tests_article"."headline",\n       '
            '"tests_article"."content",\n       '
            '"tests_article"."reporter_id"\n  '
            'FROM "tests_article"\n WHERE "tests_article"."id" = %s\n\n\n'
        )

        self.assert_msg = '4 not less than or equal to 2 queries'
        self.iter_output = '\n\nTest №1 | Queries count: 2 | Execution time: 0.743167s\n'
        self.single_output = (
            '\nQueries count: 2  |  Execution time: 0.743167s  |  Vendor: fake_vendor\n\n'
        )
        self.several_output = (
            '\n\nTests count: 1  |  Total queries count: 2  |  Total execution time: 1.48633s  |  '
            'Median time one test is: 0.743167s  |  Vendor: fake_vendor\n\n'
        )

        self.iter_dto = IterationPrintDTO(
            current_iteration=1, queries_count_per_iter=2, execution_time_per_iter=0.74316726722147
        )
        self.single_dto = SinglePrintDTO(
            queries_count=2, queries_log=queries_log, execution_time_per_iter=0.74316726722147
        )
        self.several_dto = SeveralPrintDTO(
            queries_count=2,
            queries_log=queries_log,
            current_iteration=1,
            all_execution_times=[0.74316726722147, 0.74316726722147],
        )

    def teardown_method(self, method: Callable[..., Any]) -> None:
        self.skip_colorize_sql.__exit__(None, None, None)

    @staticmethod
    def build_printer(**kwargs: Any) -> PrinterSql:
        default_printer_kwargs: dict[str, Any] = {}
        return PrinterSql(**dict(default_printer_kwargs, **kwargs))

    def test_assert_msg(self) -> None:
        obj = self.build_printer(vendor='fake_vendor', assert_q_count=2)
        data = obj.assert_msg(queries_count=4)
        assert data == self.assert_msg, repr(data)

    def test_param__verbose(self) -> None:
        with intercept_output_ctx() as ctx:
            obj = self.build_printer(
                vendor='fake_vendor', verbose=False, advanced_verb=False, queries=False
            )
            obj.print_single_sql(self.single_dto)
            output = ctx.getvalue()
            assert output == '', repr(output)

        with intercept_output_ctx() as ctx:
            obj = self.build_printer(
                vendor='fake_vendor', verbose=True, advanced_verb=False, queries=False
            )
            obj.print_single_sql(self.single_dto)
            output = ctx.getvalue()
            assert output == self.single_output, repr(output)

        with intercept_output_ctx() as ctx:
            obj = self.build_printer(
                vendor='fake_vendor', verbose=False, advanced_verb=False, queries=False
            )
            obj.print_several_sql(self.several_dto)
            output = ctx.getvalue()
            assert output == '', repr(output)

        with intercept_output_ctx() as ctx:
            obj = self.build_printer(
                vendor='fake_vendor', verbose=True, advanced_verb=False, queries=False
            )
            obj.print_several_sql(self.several_dto)
            output = ctx.getvalue()
            assert output == self.several_output, repr(output)

    def test_param__advanced_verb(self) -> None:
        with intercept_output_ctx() as ctx:
            obj = self.build_printer(
                vendor='fake_vendor', verbose=False, advanced_verb=False, queries=False
            )
            obj.iteration_print(self.iter_dto)
            output = ctx.getvalue()
            assert output == '', repr(output)

        with intercept_output_ctx() as ctx:
            obj = self.build_printer(
                vendor='fake_vendor', verbose=False, advanced_verb=True, queries=False
            )
            obj.iteration_print(self.iter_dto)
            output = ctx.getvalue()
            assert output == self.iter_output, repr(output)

    def test_param__queries(self) -> None:
        with intercept_output_ctx() as ctx:
            obj = self.build_printer(
                vendor='fake_vendor', verbose=False, advanced_verb=False, queries=False
            )
            obj.print_single_sql(self.single_dto)
            output = ctx.getvalue()
            assert output == '', repr(output)

        with intercept_output_ctx() as ctx:
            obj = self.build_printer(
                vendor='fake_vendor', verbose=False, advanced_verb=False, queries=True
            )
            obj.print_single_sql(self.single_dto)
            output = ctx.getvalue()
            assert output == self.queries_log_output, repr(output)

        with intercept_output_ctx() as ctx:
            obj = self.build_printer(
                vendor='fake_vendor', verbose=False, advanced_verb=False, queries=False
            )
            obj.print_several_sql(self.several_dto)
            output = ctx.getvalue()
            assert output == '', repr(output)

        with intercept_output_ctx() as ctx:
            obj = self.build_printer(
                vendor='fake_vendor', verbose=False, advanced_verb=False, queries=True
            )
            obj.print_several_sql(self.several_dto)
            output = ctx.getvalue()
            assert output == self.queries_log_output, repr(output)



from __future__ import annotations

import logging
from typing import Any


class CustomLogger(logging.Logger):
    def __init__(self, name: str, level: int = logging.NOTSET) -> None:
        super().__init__(name, level)

    def trace(self, msg: Any, *args: Any, **kwargs: Any) -> None:
        if is_enabled_for_trace():
            self._log(logging.DEBUG, msg, args, **kwargs)

    def dump(self, title: str, msg: Any, *args: Any, **kwargs: Any) -> None:
        if is_enabled_for_trace():
            self._log(logging.DEBUG, '--- %s ---', (title,))
            self._log(logging.DEBUG, msg, args, **kwargs)
            self._log(logging.DEBUG, '-----------------------', ())


# Registering the CustomLogger as the default logger class
logging.setLoggerClass(CustomLogger)

log: CustomLogger = logging.getLogger('capture_db_queries')
try:
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):  # type: ignore[no-redef]
        def emit(self, record: Any) -> Any:
            pass


log.addHandler(NullHandler())
log.propagate = False  # Disabling the transmission of logs higher up the chain

__all__ = ['switch_logger', 'switch_trace', 'is_enabled_for_trace']

_trace_enabled = False


def switch_logger(
    state: bool,
    handler: logging.Handler | None = None,
    formatter: logging.Formatter | None = None,
    level: str = 'DEBUG',
) -> None:
    if state:
        console_handler = handler or logging.StreamHandler()
        console_formatter = formatter or logging.Formatter(
            '%(filename)s:%(lineno)d | def %(funcName)s | %(message)s'
        )

        console_handler.setFormatter(console_formatter)
        log.addHandler(console_handler)
        log.setLevel(getattr(logging, level))
    else:
        log.addHandler(NullHandler)  # type: ignore[arg-type]


def switch_trace(traceable: bool, handler: logging.Handler | None = None, level: str = 'DEBUG') -> None:
    """
    Turn on/off the traceability.

    Parameters
    ----------
    traceable: bool
        If set to True, traceability is enabled.
    """
    global _trace_enabled
    _trace_enabled = traceable


def is_enabled_for_trace() -> bool:
    return _trace_enabled



from __future__ import annotations

from typing import TYPE_CHECKING

from typing_extensions import Self  # noqa: UP035

if TYPE_CHECKING:
    from collections.abc import Callable
    from types import TracebackType


class ContextTimer:
    def __init__(self, measure_func: Callable[..., float]) -> None:
        self.measure_func = measure_func
        self.execution_time = 0.0
        self.all_execution_times: list[float] = []
        self.exec_times_per_iter: list[float] = []

    @property
    def execution_time_per_iter(self) -> float:
        return sum(self.exec_times_per_iter)

    @property
    def queries_count_per_iter(self) -> int:
        return len(self.exec_times_per_iter)

    def __enter__(self) -> Self:
        self._start = self.measure_func()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        if exc_type is not None:
            return

        self._end = self.measure_func()
        self.execution_time = self._end - self._start
        self.all_execution_times.append(self.execution_time)
        self.exec_times_per_iter.append(self.execution_time)

    def clear_exec_times_per_iter(self) -> None:
        self.exec_times_per_iter.clear()



from __future__ import annotations

import json
import time
import traceback
from typing import TYPE_CHECKING, Any, NamedTuple

from django.utils.regex_helper import _lazy_re_compile

from ._logging import log
from .timers import ContextTimer

if TYPE_CHECKING:
    from collections.abc import Callable, Generator

    from django.db.backends.base.base import BaseDatabaseWrapper
    from django.db.backends.base.operations import BaseDatabaseOperations
    from django.db.backends.utils import CursorWrapper

    from .types import Explain, QueriesLog, Query


class BaseExecutionWrapper:
    def __init__(
        self, connection: BaseDatabaseWrapper, queries_log: QueriesLog, *args: Any, **kwargs: Any
    ) -> None:
        log.debug('')

        self.connection = connection
        self.queries_log = queries_log
        self.timer = ContextTimer(time.perf_counter)

        # Wrappers for specific databases are stored at addresses:
        # django.db.backends.sqlite3.operations.DatabaseOperations
        # django.db.backends.postgresql.operations.DatabaseOperations
        self.db_operations: BaseDatabaseOperations = self.connection.ops

    def __call__(
        self,
        execute: Callable[..., CursorWrapper],
        sql: str,
        params: tuple[Any, ...],
        many: bool,
        context: dict[str, Any],
    ) -> CursorWrapper | None:
        """
        Executes the original SQL request.
        """
        with self.timer as timer:
            try:
                result = execute(sql, params, many, context)
            except Exception as exc:
                print('Something went wrong:', exc)
                return None

        if not many:
            # Get filled SQL with params
            sql = self.db_operations.last_executed_query(context['cursor'], sql, params)

        query: Query = {'sql': sql, 'time': timer.execution_time}
        query = self.update_query(query)
        self.queries_log.append(query)

        log.trace('Location of SQL Call:\n%s', ''.join(traceback.format_stack()))

        return result

    def update_query(self, query: Query) -> Query:
        return query


class ExplainExecutionWrapper(BaseExecutionWrapper):
    """
    A class for calling EXPLAIN before each original SELECT request.
    While maintaining the full functionality of the Query Set.explain() method.

    The EXPLAIN call is not fixed in any way and does not affect the measurement results.

    https://docs.djangoproject.com/en/5.1/ref/models/querysets/#explain
    https://www.postgresql.org/docs/current/sql-explain.html
    """

    # Inspired from
    # https://www.postgresql.org/docs/current/sql-syntax-lexical.html#SQL-SYNTAX-IDENTIFIERS
    EXPLAIN_OPTIONS_PATTERN = _lazy_re_compile(r'[\w\-]+')

    class ExplainInfo(NamedTuple):
        format: str | None
        options: dict[str, Any]

    def __init__(
        self,
        connection: BaseDatabaseWrapper,
        queries_log: QueriesLog,
        explain_opts: dict[str, Any],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(connection, queries_log, *args, **kwargs)
        self.explain_info = self.build_explain_info(**explain_opts)

    def __call__(
        self,
        execute: Callable[..., CursorWrapper],
        sql: str,
        params: tuple[Any, ...],
        many: bool,
        context: dict[str, Any],
    ) -> CursorWrapper | None:
        self.__explain = self._execute_explain(sql, params, many)
        return super().__call__(execute, sql, params, many, context)

    def update_query(self, query: Query) -> Query:
        if self.__explain is not None:
            query.update({'explain': self.__explain})
        return query

    def _execute_explain(self, sql: str, params: tuple[Any, ...], many: bool) -> str | None:
        # Checking whether the request is a SELECT request
        if not sql.strip().lower().startswith('select'):
            return None

        explain_query = self.db_operations.explain_query_prefix(
            self.explain_info.format, **self.explain_info.options
        )
        explain_query = f'{explain_query} {sql}'

        try:
            raw_explain = self.__execute(explain_query, params, many)
        except Exception as exc:
            print('Something went wrong:', exc)
            return None
        else:
            return '\n'.join(self.format_explain(raw_explain))

    def __execute(self, explain_query: str, params: tuple[Any, ...], many: bool) -> Explain:
        with self.connection.cursor() as cursor:
            # you cannot call execute or executemany
            # which invokes a wrapper inside itself,
            # because it will result in duplicate call,
            # and will call __call__ of the current wrapper

            # these requests bypass wrappers
            if many:
                cursor._executemany(explain_query, params)
            else:
                cursor._execute(explain_query, params)

            return cursor.fetchall()

    def build_explain_info(self, *, format: str | None = None, **options: dict[str, Any]) -> ExplainInfo:  # noqa: A002
        """
        Validates explain options and build ExplainInfo object.
        """
        for option_name in options:
            if not self.EXPLAIN_OPTIONS_PATTERN.fullmatch(option_name) or '--' in option_name:
                raise ValueError(f'Invalid option name: {option_name!r}.')
        return self.ExplainInfo(format, options)

    def format_explain(self, result: Explain) -> Generator[str, None, None]:
        """
        Splits the explain tuple into its components and collects the final explain string from them.
        """
        nested_result = [list(result)]
        # Some backends return 1 item tuples with strings, and others return
        # tuples with integers and strings. Flatten them out into strings.
        format_ = self.explain_info.format
        output_formatter = json.dumps if format_ and format_.lower() == 'json' else str
        for row in nested_result[0]:
            if not isinstance(row, str):
                yield ' '.join(output_formatter(c) for c in row)
            else:
                yield row



from __future__ import annotations

import abc
import functools
import statistics
import time
import warnings
from collections import deque
from contextlib import contextmanager
from functools import wraps
from typing import TYPE_CHECKING, Any
from unittest.util import safe_repr

from django.core.signals import request_started
from django.db import (
    connection as db_connection,
    reset_queries,
)
from django.test.utils import CaptureQueriesContext
from typing_extensions import Self  # noqa: UP035

from ._logging import log
from .dtos import IterationPrintDTO, SeveralPrintDTO, SinglePrintDTO
from .printers import AbcPrinter, PrinterSql
from .wrappers import BaseExecutionWrapper, ExplainExecutionWrapper

if TYPE_CHECKING:
    from collections.abc import Callable, Generator
    from types import TracebackType

    from django.db.backends.base.base import BaseDatabaseWrapper
    from django.db.backends.utils import CursorWrapper

    from .types import DecoratedCallable, QueriesLog, Query, T

__all__ = ('CaptureQueries', 'capture_queries', 'ExtCaptureQueriesContext')


class AbcCapture(abc.ABC):
    def __init__(
        self,
        assert_q_count: int | None = None,
        number_runs: int = 1,
        verbose: bool = True,
        advanced_verb: bool = False,
        auto_call_func: bool = False,
        queries: bool = False,
        explain: bool = False,
        explain_opts: dict[str, Any] | None = None,
        connection: BaseDatabaseWrapper | None = None,
    ) -> None:
        log.debug('')

        self.assert_q_count = assert_q_count
        self.number_runs = number_runs
        self.verbose = verbose
        self.advanced_verb = advanced_verb
        self.auto_call_func = auto_call_func
        self.queries = queries
        self.explain = explain

        # Wrappers for specific databases are stored at addresses:
        # django.db.backends.sqlite3.base.DatabaseWrapper
        # django.db.backends.postgresql.base.DatabaseWrapper
        if connection is None:
            self.connection = db_connection
        else:
            self.connection = connection

        self.current_iteration = 0

        self.printer = self.printer_cls(
            self.connection.vendor, assert_q_count, verbose, advanced_verb, queries
        )

        self.queries_log: QueriesLog = deque(maxlen=self.connection.queries_limit)

        self.wrapper = self.wrapper_cls(
            connection=self.connection, queries_log=self.queries_log, explain_opts=explain_opts or {}
        )
        self.wrapper_ctx_manager = self.__wrap_reqs_in_wrapper()

    @property
    def printer_cls(self) -> type[AbcPrinter]:
        """Returns a class that should perform the formatting and output functions of SQL queries."""
        return PrinterSql

    @property
    def wrapper_cls(self) -> type[BaseExecutionWrapper]:
        """Returns a class that should execute all SQL queries generated by the user."""
        if self.explain:
            return ExplainExecutionWrapper
        return BaseExecutionWrapper

    @contextmanager
    def __wrap_reqs_in_wrapper(self) -> Generator[None, None, CursorWrapper]:
        """Wraps all database requests in a wrapper."""
        log.debug('')

        # https://docs.djangoproject.com/en/5.1/topics/db/instrumentation/#connection-execute-wrapper
        with self.connection.execute_wrapper(self.wrapper):
            yield

    def __enter__(self) -> Self:
        log.debug('')

        if self.number_runs > 1:
            warnings.warn(
                f'When using: {type(self).__name__} as a context manager,'
                ' the number_runs > 1 parameter is not used.',
                category=UserWarning,
                stacklevel=2,
            )

        self.wrapper_ctx_manager.__enter__()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        log.debug('')

        self.wrapper_ctx_manager.__exit__(None, None, None)

        if exc_type is not None:
            return

        self.current_iteration += 1
        queries_count = len(self)
        self.printer.print_single_sql(
            SinglePrintDTO(
                queries_count=queries_count,
                queries_log=self.queries_log,
                execution_time_per_iter=self.wrapper.timer.execution_time_per_iter,
            )
        )
        self._assert_queries_count(queries_count)

    def __iter__(self) -> Self:
        log.debug('')

        self.wrapper_ctx_manager.__enter__()
        return self

    def __next__(self) -> Self:
        log.debug('')

        if self.current_iteration > 0:
            self.printer.iteration_print(
                IterationPrintDTO(
                    current_iteration=self.current_iteration,
                    queries_count_per_iter=self.wrapper.timer.queries_count_per_iter,
                    execution_time_per_iter=self.wrapper.timer.execution_time_per_iter,
                )
            )

        self.wrapper.timer.clear_exec_times_per_iter()

        if self.current_iteration < self.number_runs:
            self.current_iteration += 1
            return self

        self.wrapper_ctx_manager.__exit__(None, None, None)

        queries_count = len(self)
        self.printer.print_several_sql(
            SeveralPrintDTO(
                queries_count=queries_count,
                queries_log=self.queries_log,
                current_iteration=self.current_iteration,
                all_execution_times=self.wrapper.timer.all_execution_times,
            )
        )
        self._assert_queries_count(queries_count)

        raise StopIteration

    def __call__(self, func: DecoratedCallable) -> DecoratedCallable | Any:
        log.debug('')

        @wraps(func)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            log.debug('')

            # The measurement consists of number_runs runs in a row
            for _ in self:
                return_value = func(*args, **kwargs)
            return return_value

        if not self.auto_call_func:
            return wrapped
        return wrapped()  # Auto-calling of the decorated function

    def __len__(self) -> int:
        log.debug('')

        return len(self.queries_log)

    def __getitem__(self, index: int) -> Query | None:
        log.debug('')

        try:
            return self.queries_log[index]
        except IndexError:
            return None

    @abc.abstractmethod
    def _assert_queries_count(self, queries_count: int) -> None:
        raise NotImplementedError


class CaptureQueries(AbcCapture):
    """
    Measures the number of database requests and
    displays detailed information about the measurements,
    validates the number of requests.

    ---

    UseCases::

        for ctx in CaptureQueries(number_runs=2, advanced_verb=True):
            response = self.client.get(url)

        OR

        @CaptureQueries(number_runs=2, advanced_verb=True)
        def test_request():
            response = self.client.get(url)

        OR

        # NOTE: The with context manager does not support multi-launch number_runs > 1
        with CaptureQueries(number_runs=1, advanced_verb=True) as ctx:
            response = self.client.get(url)

        >>> Test №1 | Queries count: 10 | Execution time: 0.04s
        >>> Test №2 | Queries count: 10 | Execution time: 0.04s
        >>> Tests count: 2  |  Total queries count: 20  |  Total execution time: 0.08s  |  Median time one test is: 0.041s  |  Vendor: sqlite

        # Example of output when using queries and explain parameters:

        for _ in CaptureQueries(advanced_verb=True, queries=True, explain=True):
            list(Reporter.objects.filter(pk=1))
            list(Article.objects.filter(pk=1))

        >>> Test №1 | Queries count: 2 | Execution time: 0.22s
        >>>
        >>>
        >>> №[1] time=[0.109] explain=['2 0 0 SEARCH TABLE tests_reporter USING INTEGER PRIMARY KEY (rowid=?)']
        >>> SELECT "tests_reporter"."id",
        >>>     "tests_reporter"."full_name"
        >>> FROM "tests_reporter"
        >>> WHERE "tests_reporter"."id" = %s
        >>>
        >>>
        >>> №[2] time=[0.109] explain=['2 0 0 SEARCH TABLE tests_article USING INTEGER PRIMARY KEY (rowid=?)']
        >>> SELECT "tests_article"."id",
        >>>     "tests_article"."pub_date",
        >>>     "tests_article"."headline",
        >>>     "tests_article"."content",
        >>>     "tests_article"."reporter_id"
        >>> FROM "tests_article"
        >>> WHERE "tests_article"."id" = %s
        >>>
        >>>
        >>> Tests count: 1  |  Total queries count: 2  |  Total execution time: 0.22s  |  Median time one test is: 0.109s  |  Vendor: sqlite

    ---

    - assert_q_count: The expected number of database requests otherwise an
     "AssertionError: N not less than or equal to N queries"
    - number_runs: The number of runs of the test function / test cycle
    - verbose: Displaying the final results of the test measurements
    - advanced_verb: Displaying the result of each test measurement
    - auto_call_func: Autorun of the decorated function (without arguments)
    - queries: Displaying raw SQL queries to the database
    - explain: Displaying additional information about each request (has no effect on the orig. requests)
    - explain_opts: Parameters for explain, find out more in the documentation for your DBMS
    - connection: Connecting to your database, by default: django.db.connection

    ---

    Tests count: Total number of test measurements
    Total queries count: The total number of database requests
     within the test function within all measurements
    Total execution time: Total time for executing database queries
     inside the test function within all measurements
    Median time one test is: Average execution time of one test measurement
    Vendor: The database under test
    """  # noqa: E501

    def _assert_queries_count(self, queries_count: int) -> None:
        log.debug('')

        if self.assert_q_count is not None:
            assert queries_count <= self.assert_q_count, self.printer.assert_msg(queries_count)


def capture_queries(
    assert_q_count: int | None = None,
    number_runs: int = 1,
    verbose: bool = True,
    advanced_verb: bool = False,
    queries: bool = False,
) -> Callable[[DecoratedCallable], None]:
    """
    Замеряет количество запросов к бд внутри тела тестовой функции,
    выводит подробную информацию о замерах,
    валидирует количество запросов.

    Вызов тестовой функции `_` произойдёт автоматически.

    UseCase::

        @capture_queries(number_runs=2, advanced_verb=True)
        def _():
            response = self.client.get(url)

        >>> Test №1 | Queries count: 10 | Execution time: 0.04s
        >>> Test №2 | Queries count: 10 | Execution time: 0.04s
        >>> Tests count: 2  |  Total queries count: 20  |  Total execution time: 0.08s  |  Median time one test is: 0.041s

    - assert_q_count: Ожидаемое количество запросов к БД иначе
     "AssertionError: N not less than or equal to N queries"
    - number_runs: Количество запусков тестовой функции `_`
    - verbose: Отображение финальных результатов тестовых замеров
    - advanced_verb: Отображение результа каждого тестового замера
    - queries: Отображение сырых SQL запросов к БД

    Tests count: Общее количество тестовых замеров
    Total queries count: Общее количество запросов к БД внутри тестовой функции в рамках всех замеров
    Total execution time: Общее время выполнения запросов к БД
     внутри тестовой функции в рамках всех замеров
    Median time one test is: Среднее время выполнения одного тестового замера
    """  # noqa: E501

    def _wrapper(func: DecoratedCallable) -> None:
        """"""

        @functools.wraps(func)
        def spoof_func(*args: Any, **kwargs: Any) -> None:
            """"""
            warnings.warn(
                'Декоратор @capture_queries устарел и будет удален в будущих версиях. '
                'Используйте класс @CaptureQueries вместо него.',
                category=DeprecationWarning,
                stacklevel=2,
            )

            force_debug_cursor = db_connection.force_debug_cursor
            db_connection.force_debug_cursor = True
            # Run any initialization queries if needed so that they won't be
            # included as part of the count.
            db_connection.ensure_connection()
            initial_queries = len(db_connection.queries_log)
            final_queries = None
            request_started.disconnect(reset_queries)

            all_execution_times = []
            for i in range(1, number_runs + 1):  # замер состоит из number_runs прогонов подряд
                start_queries = len(db_connection.queries)
                start = time.perf_counter()

                return_value = func(*args, **kwargs)

                end = time.perf_counter()
                end_queries = len(db_connection.queries)
                execution_time = end - start

                all_execution_times.append(execution_time)
                if advanced_verb:
                    print(
                        f'Test №{i} | '
                        f'Queries count: {end_queries - start_queries} | '
                        f'Execution time: {execution_time:.2f}s'
                    )

            db_connection.force_debug_cursor = force_debug_cursor
            request_started.connect(reset_queries)
            final_queries = len(db_connection.queries_log)

            if queries:
                for query in db_connection.queries[initial_queries:final_queries]:
                    print(query, end='\n\n')

            if verbose:
                print(
                    f'Tests count: {i}  |  '
                    f'Total queries count: {final_queries}  |  '
                    f'Total execution time: {sum(all_execution_times):.2f}s  |  '
                    f'Median time one test is: {statistics.median(all_execution_times):.3f}s'
                )

            if assert_q_count is not None:
                standard_msg = (
                    f'{safe_repr(final_queries)} not less than or equal to '
                    f'{safe_repr(assert_q_count)} queries'
                )
                assert final_queries <= assert_q_count, standard_msg

            return return_value

        spoof_func()  # автовызов декорируемой функции

    return _wrapper


class ExtCaptureQueriesContext(CaptureQueriesContext):
    """
    Замеряет количество запросов к бд внутри тела контекстного менеджера,
    выводит подробную информацию о замере,
    валидирует количество запросов.

    UseCase::

        with ExtCaptureQueriesContext():
            response = self.client.get(url)

        >>> Queries count: 164  |  Execution time: 0.923s

    - assert_q_count: Ожидаемое количество запросов к БД иначе
     "AssertionError: N not less than or equal to N queries"
    - verbose: Отображение финальных результатов тестового замера
    - queries: Отображение сырых SQL запросов к БД

    Queries count: Количество запросов к БД внутри контекстного менеджера
    Execution time: Время выполнения запросов к БД внутри контекстного менеджера
    """

    def __init__(
        self, assert_q_count: int | None = None, verbose: bool = True, queries: bool = False
    ) -> None:
        warnings.warn(
            'Класс контекстного менеджера with ExtCaptureQueriesContext '
            'устарел и будет удален в будущих версиях. '
            'Используйте класс @CaptureQueries вместо него.',
            category=DeprecationWarning,
            stacklevel=2,
        )

        super().__init__(db_connection)

        self.assert_q_count = assert_q_count
        self.verbose = verbose
        self.queries = queries

    def __enter__(self: T) -> T:
        self.start = time.perf_counter()
        return super().__enter__()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        super().__exit__(exc_type, exc_value, traceback)

        end = time.perf_counter()
        execution_time = end - self.start

        if self.queries:
            for query in self.captured_queries:
                print(query, end='\n\n')

        if self.verbose:
            print(f'Queries count: {self.final_queries}  |  Execution time: {execution_time:.3f}s')

        if self.assert_q_count is not None:
            standard_msg = (
                f'{safe_repr(self.final_queries)} not less than or equal to '
                f'{safe_repr(self.assert_q_count)} queries'
            )
            assert self.final_queries <= self.assert_q_count, standard_msg



from django.db import models


class Reporter(models.Model):
    full_name = models.CharField(max_length=70)

    def __str__(self) -> str:
        return self.full_name


class Article(models.Model):
    pub_date = models.DateField()
    headline = models.CharField(max_length=200)
    content = models.TextField()
    reporter = models.ForeignKey(Reporter, on_delete=models.CASCADE)

    def __str__(self) -> str:
        return self.headline
